#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

int counter;
const int MAX_COUNT = 1e6;

int lock_var = 0;
// http://www.ibiblio.org/gferg/ldp/GCC-Inline-Assembly-HOWTO.html#s5

void lock_tas() {
    // Exchange the local `tmp` variable atomically with the value in `lock_var`.
    // If the after the exchange the value of `tmp` is zero, then the thread
    // managed to acquire the lock variable and therefore can continue into
    // the critical region. Otherwise, the value of `tmp` will be not zero
    // and the thread will start another attempt to acquire the lock.
    register unsigned char tmp = 1;
    while (tmp != 0) {
     __asm__ __volatile__(
        "xchg %0,%1"
       :"+q" (tmp), "+m" (lock_var)  /* output operands */
       :                              /* input operands */
       : );                           /* clobbered operands */
    }

    // At this point, the thread managed to acquire the lock and will proceed
    // into the critical region.
}

void unlock_tas() {
    // The following "mfence" is not required to make our example work. However
    // in more complex setups, where there are more memory operations after the
    // call to unlock this "mfence" is necessary.
    __asm__ __volatile__(
        "mfence"
       :
       :
       // The "mfence" command can affect all memory locations, so tell the
       // __asm__ about the effect on the total memroy system.
       :"memory");

    // Reset the lock variable such that the lock can be acquired again.
    lock_var = 0;
}

void incr_func( void *ptr ) {
     printf("Hello from thread\n");

     for (int i = 0; i < MAX_COUNT; i++) {
         lock_tas();

         counter += 1;

         unlock_tas();
     }
}

int main() {
    // Reset the global counter variable.
    counter = 0;

    int err;
    pthread_t thread1, thread2;

    // Create two threads that both increment the counter variable 1e6 times.
    if((err = pthread_create(&thread1, NULL, (void *) &incr_func, NULL))) {
        fprintf(stderr, "Error - pthread_create() return code: %d\n", err);
        exit(EXIT_FAILURE);
    }

    if((err = pthread_create(&thread2, NULL, (void *) &incr_func, NULL))) {
        fprintf(stderr, "Error - pthread_create() return code: %d\n", err);
        exit(EXIT_FAILURE);
    }

    // Wait until both threads have finished.
    if (err = pthread_join(thread1, NULL)) {
        fprintf(stderr, "Error - pthread_join(thread1) return code: %d\n", err);
        exit(EXIT_FAILURE);
    }
    if (err = pthread_join(thread2, NULL)) {
        fprintf(stderr, "Error - pthread_join(thread2) return code: %d\n", err);
        exit(EXIT_FAILURE);
    }
    printf("\n");

    // Print the final counter variable
    printf("Counter:  %d\n", counter);
    printf("Expected: %d\n", 2 * MAX_COUNT);

    return 0;
}


