#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>

void TERMHandler(int sig) {
    printf("CHILD: Got SIGTERM signal but ignoring it for now.\n");

    // Adjust the signal handler such that the next time the `SIGTERM` signal
    // is recieved the default action is performed instead of this custom
    // event handler. In particular, falling back to the default handler will
    // cause the process to get killed.
    if (signal(sig, SIG_DFL) == SIG_ERR) {
        fprintf(stderr, "can't setup signal handler, error %d\n", errno);
        exit(EXIT_FAILURE);
    }
}

int main(int argc, char** argv) {
    pid_t pid = fork();

    if (pid == -1) {
       fprintf(stderr, "can't fork, error %d\n", errno);
       exit(EXIT_FAILURE);
    } else if (pid == 0) {
        int counter = 0;

        // Setup a signal handler for the `SIGTERM` signal. This allows to
        // ignore the `SIGTERM` the first time.
        if (signal(SIGTERM, TERMHandler) == SIG_ERR) {
            fprintf(stderr, "can't setup signal handler, error %d\n", errno);
            exit(EXIT_FAILURE);
        }

        // Spinning in a while loop. This process is killed by the parent.
        while (1 /* true */) {
            printf("CHILD counter: %d\n", counter);
            counter += 1;
            sleep(1);
        }
    } else {
        sleep(2);

        // Pause the child process.
        printf("PARENT: Pause the child.\n");
        if (kill(pid, SIGTSTP)) {
            fprintf(stderr, "can't send signal to child, error %d\n", errno);
            exit(EXIT_FAILURE);
        }

        sleep(2);

        // Continue the child process again.
        printf("PARENT: Continue the child.\n");
        if (kill(pid, SIGCONT)) {
            fprintf(stderr, "can't send signal to child, error %d\n", errno);
            exit(EXIT_FAILURE);
        }

        sleep(2);

        // Try to terminate the child process once.
        printf("PARENT: Send SIGTERM to child.\n");
        if (kill(pid, SIGTERM)) {
            fprintf(stderr, "can't send signal to child, error %d\n", errno);
            exit(EXIT_FAILURE);
        }

        sleep(2);

        // Try to terminate the child process once.
        printf("PARENT: Send SIGTERM to child again.\n");
        if (kill(pid, SIGTERM)) {
            fprintf(stderr, "can't send signal to child, error %d\n", errno);
            exit(EXIT_FAILURE);
        }
    }

    return 0;
}
